==> picodata-23.06.0.73.tar.xz: picodata-23.06.0.73/src/on_shutdown.rs <==
use std::time::{Duration, Instant};

use ::tarantool::fiber;

use crate::has_grades;
use crate::instance::grade::TargetGradeVariant;
use crate::rpc;
use crate::storage::ClusterwideSpaceId;
use crate::tlog;
use crate::traft;
use crate::traft::node;
use crate::unwrap_ok_or;

pub async fn callback() {
    // 1. Try setting target grade Offline in a separate fiber
    tlog!(Info, "trying to shutdown gracefully ...");
    let go_offline = fiber::Builder::new()
        .name("go_offline")
        .func(|| match go_offline() {
            Ok(()) => tlog!(Info, "target grade set Offline"),
            Err(e) => tlog!(Error, "failed setting target grade Offline: {e}"),
        });
    std::mem::forget(go_offline.start());

    // 2. Meanwhile, wait until either it succeeds or there is no quorum.
    let node = node::global().unwrap();
    let raft_id = node.raft_id();
    let mut instances_watcher = node.storage_watcher(ClusterwideSpaceId::Instance);
    loop {
        let me = unwrap_ok_or!(
            node.storage.instances.get(&raft_id),
            Err(e) => {
                tlog!(Error, "{e}");
                break;
            }
        );

        if has_grades!(me, Offline -> *) {
            tlog!(Info, "graceful shutdown succeeded");

            // Dirty hack. Wait a little bit more before actually
            // shutting down. Raft commit index is a local value. Other
            // nodes may still be unaware that `me.current_grade` is
            // commmitted. Give them some more time to communicate.
            fiber::sleep(Duration::from_millis(100));
            break;
        }

        let voters = node.raft_storage.voters().expect("failed reading voters");

        let quorum = voters.len() / 2 + 1;
        let voters_alive = voters
            .iter()
            .filter_map(|raft_id| node.storage.instances.get(raft_id).ok())
            .filter(|instance| has_grades!(instance, Online -> *))
            .count();

        if voters_alive < quorum {
            tlog!(Info, "giving up, there is no quorum");
            break;
        }

        if let Err(e) = instances_watcher.changed().await {
            tlog!(Warning, "failed to shutdown gracefully: {e}");
        }
    }
}

fn go_offline() -> traft::Result<()> {
    let node = node::global()?;
    let raft_id = node.raft_id();

    let instance = node.storage.instances.get(&raft_id)?;
    let cluster_id = node.raft_storage.cluster_id()?;

    let req = rpc::update_instance::Request::new(instance.instance_id, cluster_id)
        .with_target_grade(TargetGradeVariant::Offline);

    loop {
        let now = Instant::now();
        let wait_before_retry = Duration::from_millis(300);

        match node.handle_update_instance_request_and_wait(req.clone(), wait_before_retry) {
            Ok(_) => break Ok(()),
            Err(e) => {
                tlog!(Warning,
                    "failed setting target grade Offline: {e}, retrying ...";
                );
                fiber::sleep(wait_before_retry.saturating_sub(now.elapsed()));
                continue;
            }
        };
    }
}

==> picodata-23.06.0.275.tar.xz: picodata-23.06.0.275/src/on_shutdown.rs <==
use std::time::Duration;

use ::tarantool::fiber;

use crate::has_grades;
use crate::tlog;
use crate::traft::event;
use crate::traft::node;
use crate::unwrap_ok_or;

use picodata_sdk::*;

pub async fn callback(plugin_list: &'static [Plugin]) {
    let node = node::global().unwrap();

    // 1. Wake up the sentinel so it starts trying to set target grade Offline.
    node.sentinel_loop.on_shut_down();
    plugin_list.iter().for_each(|plugin| plugin.stop());
    fiber::reschedule();

    // 2. Meanwhile, wait until either it succeeds or there is no quorum.
    let raft_id = node.raft_id();
    loop {
        let me = unwrap_ok_or!(
            node.storage.instances.get(&raft_id),
            Err(e) => {
                tlog!(Error, "{e}");
                break;
            }
        );

        if has_grades!(me, Offline -> *) {
            tlog!(Info, "graceful shutdown succeeded");

            // Dirty hack. Wait a little bit more before actually
            // shutting down. Raft commit index is a local value. Other
            // nodes may still be unaware that `me.current_grade` is
            // commmitted. Give them some more time to communicate.
            fiber::sleep(Duration::from_millis(100));
            break;
        }

        let voters = node.raft_storage.voters().expect("failed reading voters");

        let quorum = voters.len() / 2 + 1;
        let voters_alive = voters
            .iter()
            .filter_map(|raft_id| node.storage.instances.get(raft_id).ok())
            .filter(|instance| has_grades!(instance, Online -> *))
            .count();

        if voters_alive < quorum {
            tlog!(Info, "giving up, there is no quorum");
            break;
        }

        if let Err(e) = event::wait_timeout(event::Event::EntryApplied, Duration::MAX) {
            tlog!(Warning, "failed to shutdown gracefully: {e}");
        }
    }
}
